from __future__ import annotations

import abc
import asyncio
import concurrent.futures
import contextvars
import datetime
import functools
import logging
import typing
from typing import (
    Optional,
    Any,
    Iterator,
    AsyncIterator,
    Callable,
    Iterable,
    Union,
    Coroutine,
)
from dataclasses import dataclass

import grpc
from google.protobuf.message import Message
from google.protobuf.duration_pb2 import Duration as ProtoDuration
from google.protobuf.timestamp_pb2 import Timestamp as ProtoTimeStamp

from ...driver import Driver
from ...aio.driver import Driver as DriverIO

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ..v4.protos import ydb_topic_pb2, ydb_issue_message_pb2
else:
    from ..common.protos import ydb_topic_pb2, ydb_issue_message_pb2

from ... import issues, connection
from ...settings import BaseRequestSettings
from ..._constants import DEFAULT_LONG_STREAM_TIMEOUT

logger = logging.getLogger(__name__)


class IFromProto(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def from_proto(msg: Message) -> Any:
        ...


class IFromProtoWithProtoType(IFromProto):
    @staticmethod
    @abc.abstractmethod
    def empty_proto_message() -> Message:
        ...


class IToProto(abc.ABC):
    @abc.abstractmethod
    def to_proto(self) -> Message:
        ...


class IFromPublic(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def from_public(o: typing.Any) -> typing.Any:
        ...


class IToPublic(abc.ABC):
    @abc.abstractmethod
    def to_public(self) -> typing.Any:
        ...


class UnknownGrpcMessageError(issues.Error):
    pass


_stop_grpc_connection_marker = object()


class QueueToIteratorAsyncIO:
    __slots__ = ("_queue",)

    def __init__(self, q: asyncio.Queue):
        self._queue = q

    def __aiter__(self):
        return self

    async def __anext__(self):
        item = await self._queue.get()
        if item is _stop_grpc_connection_marker:
            raise StopAsyncIteration()
        return item


class AsyncQueueToSyncIteratorAsyncIO:
    __slots__ = (
        "_loop",
        "_queue",
    )
    _queue: asyncio.Queue

    def __init__(self, q: asyncio.Queue):
        self._loop = asyncio.get_running_loop()
        self._queue = q

    def __iter__(self):
        return self

    def __next__(self):
        item = asyncio.run_coroutine_threadsafe(self._queue.get(), self._loop).result()
        if item is _stop_grpc_connection_marker:
            raise StopIteration()
        return item


class SyncToAsyncIterator:
    def __init__(self, sync_iterator: Iterator, executor: concurrent.futures.Executor):
        self._sync_iterator = sync_iterator
        self._executor = executor

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            res = await to_thread(self._sync_iterator.__next__, executor=self._executor)
            return res
        except StopIteration:
            raise StopAsyncIteration()


class IGrpcWrapperAsyncIO(abc.ABC):
    @abc.abstractmethod
    async def receive(self, timeout: Optional[int] = None) -> Any:
        ...

    @abc.abstractmethod
    def write(self, wrap_message: IToProto):
        ...

    @abc.abstractmethod
    def close(self):
        ...


SupportedDriverType = Union[Driver, DriverIO]


class GrpcWrapperAsyncIO(IGrpcWrapperAsyncIO):
    from_client_grpc: asyncio.Queue
    from_server_grpc: AsyncIterator
    convert_server_grpc_to_wrapper: Callable[[Any], Any]
    _connection_state: str
    _stream_call: Optional[Union[grpc.aio.StreamStreamCall, "grpc._channel._MultiThreadedRendezvous"]]
    _wait_executor: Optional[concurrent.futures.ThreadPoolExecutor]

    def __init__(self, convert_server_grpc_to_wrapper):
        self.from_client_grpc = asyncio.Queue()
        self.convert_server_grpc_to_wrapper = convert_server_grpc_to_wrapper
        self._connection_state = "new"
        self._stream_call = None
        self._wait_executor = None

        self._stream_settings: BaseRequestSettings = (
            BaseRequestSettings()
            .with_operation_timeout(DEFAULT_LONG_STREAM_TIMEOUT)
            .with_cancel_after(DEFAULT_LONG_STREAM_TIMEOUT)
            .with_timeout(DEFAULT_LONG_STREAM_TIMEOUT)
        )

    def __del__(self):
        self._clean_executor(wait=False)

    async def start(self, driver: SupportedDriverType, stub, method):
        if asyncio.iscoroutinefunction(driver.__call__):
            await self._start_asyncio_driver(driver, stub, method)
        else:
            await self._start_sync_driver(driver, stub, method)
        self._connection_state = "started"

    def close(self):
        self.from_client_grpc.put_nowait(_stop_grpc_connection_marker)
        if self._stream_call:
            self._stream_call.cancel()

        self._clean_executor(wait=True)

    def _clean_executor(self, wait: bool):
        if self._wait_executor:
            self._wait_executor.shutdown(wait)

    async def _start_asyncio_driver(self, driver: DriverIO, stub, method):
        requests_iterator = QueueToIteratorAsyncIO(self.from_client_grpc)
        stream_call = await driver(
            requests_iterator,
            stub,
            method,
            settings=self._stream_settings,
        )
        self._stream_call = stream_call
        self.from_server_grpc = stream_call.__aiter__()

    async def _start_sync_driver(self, driver: Driver, stub, method):
        requests_iterator = AsyncQueueToSyncIteratorAsyncIO(self.from_client_grpc)
        self._wait_executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        stream_call = await to_thread(
            driver,
            requests_iterator,
            stub,
            method,
            executor=self._wait_executor,
            settings=self._stream_settings,
        )
        self._stream_call = stream_call
        self.from_server_grpc = SyncToAsyncIterator(stream_call.__iter__(), self._wait_executor)

    async def receive(self, timeout: Optional[int] = None, is_coordination_calls=False) -> Any:
        # todo handle grpc exceptions and convert it to internal exceptions
        try:
            if timeout is None:
                grpc_message = await self.from_server_grpc.__anext__()
            else:

                async def get_response():
                    return await self.from_server_grpc.__anext__()

                grpc_message = await asyncio.wait_for(get_response(), timeout)

        except (grpc.RpcError, grpc.aio.AioRpcError) as e:
            raise connection._rpc_error_handler(self._connection_state, e)

        if not is_coordination_calls:
            issues._process_response(grpc_message)

        if self._connection_state != "has_received_messages":
            self._connection_state = "has_received_messages"

        # print("rekby, grpc, received", grpc_message)
        return self.convert_server_grpc_to_wrapper(grpc_message)

    def write(self, wrap_message: IToProto):
        grpc_message = wrap_message.to_proto()
        # print("rekby, grpc, send", grpc_message)
        self.from_client_grpc.put_nowait(grpc_message)


@dataclass(init=False)
class ServerStatus(IFromProto):
    __slots__ = ("_grpc_status_code", "_issues")

    def __init__(
        self,
        status: issues.StatusCode,
        issues: Iterable[Any],
    ):
        self.status = status
        self.issues = issues

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def from_proto(
        msg: Union[
            ydb_topic_pb2.StreamReadMessage.FromServer,
            ydb_topic_pb2.StreamWriteMessage.FromServer,
        ]
    ) -> "ServerStatus":
        return ServerStatus(msg.status, msg.issues)

    def is_success(self) -> bool:
        return self.status == issues.StatusCode.SUCCESS

    @classmethod
    def issue_to_str(cls, issue: ydb_issue_message_pb2.IssueMessage):
        res = """code: %s message: "%s" """ % (issue.issue_code, issue.message)
        if len(issue.issues) > 0:
            d = ", "
            res += d + d.join(str(sub_issue) for sub_issue in issue.issues)
        return res


def callback_from_asyncio(callback: Union[Callable, Coroutine]) -> [asyncio.Future, asyncio.Task]:
    loop = asyncio.get_running_loop()

    if asyncio.iscoroutinefunction(callback):
        return loop.create_task(callback())
    else:
        return loop.run_in_executor(None, callback)


async def to_thread(func, *args, executor: Optional[concurrent.futures.Executor], **kwargs):
    """Asynchronously run function *func* in a separate thread.

    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    Return a coroutine that can be awaited to get the eventual result of *func*.

    copy to_thread from 3.10
    """

    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(executor, func_call)


def proto_duration_from_timedelta(t: Optional[datetime.timedelta]) -> Optional[ProtoDuration]:
    if t is None:
        return None

    res = ProtoDuration()
    res.FromTimedelta(t)
    return res


def proto_timestamp_from_datetime(t: Optional[datetime.datetime]) -> Optional[ProtoTimeStamp]:
    if t is None:
        return None

    res = ProtoTimeStamp()
    res.FromDatetime(t)
    return res


def datetime_from_proto_timestamp(
    ts: Optional[ProtoTimeStamp],
) -> Optional[datetime.datetime]:
    if ts is None:
        return None
    return ts.ToDatetime()


def timedelta_from_proto_duration(
    d: Optional[ProtoDuration],
) -> Optional[datetime.timedelta]:
    if d is None:
        return None
    return d.ToTimedelta()



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/ydb_object_storage_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.draft.protos import ydb_object_storage_pb2 as draft_dot_protos_dot_ydb__object__storage__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!draft/ydb_object_storage_v1.proto\x12\x14Ydb.ObjectStorage.V1\x1a%draft/protos/ydb_object_storage.proto2e\n\x14ObjectStorageService\x12M\n\x04List\x12!.Ydb.ObjectStorage.ListingRequest\x1a\".Ydb.ObjectStorage.ListingResponseBl\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'draft.ydb_object_storage_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1'
  _OBJECTSTORAGESERVICE._serialized_start=98
  _OBJECTSTORAGESERVICE._serialized_end=199
# @@protoc_insertion_point(module_scope)



import json

import pytest
import ydb
import time
from concurrent import futures

from typing import Optional

from ydb import QueryExplainResultFormat
from ydb.query.pool import QuerySessionPool
from ydb.query.session import QuerySession
from ydb.query.transaction import QueryTxContext


class TestQuerySessionPool:
    def test_checkout_provides_created_session(self, pool: QuerySessionPool):
        with pool.checkout() as session:
            assert session.is_active

    def test_oneshot_query_normal(self, pool: QuerySessionPool):
        res = pool.execute_with_retries("select 1;")
        assert len(res) == 1

    def test_oneshot_query_result_set_index(self, pool: QuerySessionPool):
        res = pool.execute_with_retries("select 1; select 2; select 3")
        assert len(res) == 3
        indexes = [result_set.index for result_set in res]
        assert indexes == [0, 1, 2]

    def test_oneshot_ddl_query(self, pool: QuerySessionPool):
        pool.execute_with_retries("create table Queen(key UInt64, PRIMARY KEY (key));")
        pool.execute_with_retries("drop table Queen;")

    def test_oneshot_query_raises(self, pool: QuerySessionPool):
        with pytest.raises(ydb.GenericError):
            pool.execute_with_retries("Is this the real life? Is this just fantasy?")

    def test_retry_op_uses_created_session(self, pool: QuerySessionPool):
        def callee(session: QuerySession):
            assert session.is_active

        pool.retry_operation_sync(callee)

    def test_retry_op_normal(self, pool: QuerySessionPool):
        def callee(session: QuerySession):
            with session.transaction() as tx:
                iterator = tx.execute("select 1;", commit_tx=True)
                return [result_set for result_set in iterator]

        res = pool.retry_operation_sync(callee)
        assert len(res) == 1

    def test_retry_op_raises(self, pool: QuerySessionPool):
        class CustomException(Exception):
            pass

        def callee(session: QuerySession):
            raise CustomException()

        with pytest.raises(CustomException):
            pool.retry_operation_sync(callee)

    @pytest.mark.parametrize(
        "tx_mode",
        [
            (None),
            (ydb.QuerySerializableReadWrite()),
            (ydb.QuerySnapshotReadOnly()),
            (ydb.QuerySnapshotReadWrite()),
            (ydb.QueryOnlineReadOnly()),
            (ydb.QueryStaleReadOnly()),
        ],
    )
    def test_retry_tx_normal(self, pool: QuerySessionPool, tx_mode: Optional[ydb.BaseQueryTxMode]):
        retry_no = 0

        def callee(tx: QueryTxContext):
            nonlocal retry_no
            if retry_no < 2:
                retry_no += 1
                raise ydb.Unavailable("Fake fast backoff error")
            result_stream = tx.execute("SELECT 1")
            return [result_set for result_set in result_stream]

        result = pool.retry_tx_sync(callee=callee, tx_mode=tx_mode)
        assert len(result) == 1
        assert retry_no == 2

    def test_retry_tx_raises(self, pool: QuerySessionPool):
        class CustomException(Exception):
            pass

        def callee(tx: QueryTxContext):
            raise CustomException()

        with pytest.raises(CustomException):
            pool.retry_tx_sync(callee)

    def test_pool_size_limit_logic(self, pool: QuerySessionPool):
        target_size = 5
        pool._size = target_size
        ids = set()

        for i in range(1, target_size + 1):
            session = pool.acquire(timeout=0.1)
            assert pool._current_size == i
            assert session.session_id not in ids
            ids.add(session.session_id)

        with pytest.raises(ydb.SessionPoolEmpty):
            pool.acquire(timeout=0.1)

        last_id = session.session_id
        pool.release(session)

        session = pool.acquire(timeout=0.1)
        assert session.session_id == last_id
        assert pool._current_size == target_size

    def test_checkout_do_not_increase_size(self, pool: QuerySessionPool):
        session_id = None
        for _ in range(10):
            with pool.checkout() as session:
                if session_id is None:
                    session_id = session.session_id
                assert pool._current_size == 1
                assert session_id == session.session_id

    def test_pool_recreates_bad_sessions(self, pool: QuerySessionPool):
        with pool.checkout() as session:
            session_id = session.session_id
            session.delete()

        with pool.checkout() as session:
            assert session_id != session.session_id
            assert pool._current_size == 1

    def test_acquire_from_closed_pool_raises(self, pool: QuerySessionPool):
        pool.stop()
        with pytest.raises(ydb.SessionPoolClosed):
            pool.acquire(1)

    def test_no_session_leak(self, driver_sync, docker_project):
        pool = ydb.QuerySessionPool(driver_sync, 1)
        docker_project.stop()
        try:
            pool.acquire(timeout=0.1)
        except ydb.Error:
            pass
        assert pool._current_size == 0

        docker_project.start()
        pool.stop()

    def test_execute_with_retries_async(self, pool: QuerySessionPool):
        fut = pool.execute_with_retries_async("select 1;")
        res = fut.result()
        assert len(res) == 1

    def test_retry_operation_async(self, pool: QuerySessionPool):
        def callee(session: QuerySession):
            with session.transaction() as tx:
                iterator = tx.execute("select 1;", commit_tx=True)
                return [result_set for result_set in iterator]

        fut = pool.retry_operation_async(callee)
        res = fut.result()
        assert len(res) == 1

    def test_retry_tx_async(self, pool: QuerySessionPool):
        retry_no = 0

        def callee(tx: QueryTxContext):
            nonlocal retry_no
            if retry_no < 2:
                retry_no += 1
                raise ydb.Unavailable("Fake fast backoff error")
            result_stream = tx.execute("SELECT 1")
            return [result_set for result_set in result_stream]

        result = pool.retry_tx_async(callee=callee).result()
        assert len(result) == 1
        assert retry_no == 2

    def test_execute_with_retries_async_many_calls(self, pool: QuerySessionPool):
        futs = [pool.execute_with_retries_async("select 1;") for _ in range(10)]
        results = [f.result() for f in futures.as_completed(futs)]
        assert all(len(r) == 1 for r in results)

    def test_future_waits_on_stop(self, pool: QuerySessionPool):
        def callee(session: QuerySession):
            time.sleep(0.1)
            with session.transaction() as tx:
                it = tx.execute("select 1;", commit_tx=True)
                return [rs for rs in it]

        fut = pool.retry_operation_async(callee)
        pool.stop()
        assert fut.done()
        assert len(fut.result()) == 1

    def test_async_methods_after_stop_raise(self, pool: QuerySessionPool):
        pool.stop()
        with pytest.raises(ydb.SessionPoolClosed):
            pool.execute_with_retries_async("select 1;")

    def test_explain_with_retries(self, pool: QuerySessionPool):
        pool.execute_with_retries("DROP TABLE IF EXISTS test_explain")
        pool.execute_with_retries("CREATE TABLE test_explain (id Int64, PRIMARY KEY (id))")
        try:

            plan = pool.explain_with_retries("SELECT * FROM test_explain", result_format=QueryExplainResultFormat.STR)
            isinstance(plan, str)
            assert "FullScan" in plan

            plan = pool.explain_with_retries("SELECT * FROM test_explain", result_format=QueryExplainResultFormat.DICT)
            assert isinstance(plan, dict)

            plan_string = json.dumps(plan)
            assert "FullScan" in plan_string

            plan = pool.explain_with_retries(
                "SELECT * FROM test_explain WHERE id = $id",
                {"$id": 1},
                result_format=ydb.QueryExplainResultFormat.DICT,
            )
            plan_string = json.dumps(plan)
            assert "Lookup" in plan_string
        finally:
            pool.execute_with_retries("DROP TABLE test_explain")



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_table_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_table_pb2 as protos_dot_ydb__table__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12ydb_table_v1.proto\x12\x0cYdb.Table.V1\x1a\x16protos/ydb_table.proto2\xe9\x0e\n\x0cTableService\x12R\n\rCreateSession\x12\x1f.Ydb.Table.CreateSessionRequest\x1a .Ydb.Table.CreateSessionResponse\x12R\n\rDeleteSession\x12\x1f.Ydb.Table.DeleteSessionRequest\x1a .Ydb.Table.DeleteSessionResponse\x12\x46\n\tKeepAlive\x12\x1b.Ydb.Table.KeepAliveRequest\x1a\x1c.Ydb.Table.KeepAliveResponse\x12L\n\x0b\x43reateTable\x12\x1d.Ydb.Table.CreateTableRequest\x1a\x1e.Ydb.Table.CreateTableResponse\x12\x46\n\tDropTable\x12\x1b.Ydb.Table.DropTableRequest\x1a\x1c.Ydb.Table.DropTableResponse\x12I\n\nAlterTable\x12\x1c.Ydb.Table.AlterTableRequest\x1a\x1d.Ydb.Table.AlterTableResponse\x12\x46\n\tCopyTable\x12\x1b.Ydb.Table.CopyTableRequest\x1a\x1c.Ydb.Table.CopyTableResponse\x12I\n\nCopyTables\x12\x1c.Ydb.Table.CopyTablesRequest\x1a\x1d.Ydb.Table.CopyTablesResponse\x12O\n\x0cRenameTables\x12\x1e.Ydb.Table.RenameTablesRequest\x1a\x1f.Ydb.Table.RenameTablesResponse\x12R\n\rDescribeTable\x12\x1f.Ydb.Table.DescribeTableRequest\x1a .Ydb.Table.DescribeTableResponse\x12[\n\x10\x45xplainDataQuery\x12\".Ydb.Table.ExplainDataQueryRequest\x1a#.Ydb.Table.ExplainDataQueryResponse\x12[\n\x10PrepareDataQuery\x12\".Ydb.Table.PrepareDataQueryRequest\x1a#.Ydb.Table.PrepareDataQueryResponse\x12[\n\x10\x45xecuteDataQuery\x12\".Ydb.Table.ExecuteDataQueryRequest\x1a#.Ydb.Table.ExecuteDataQueryResponse\x12\x61\n\x12\x45xecuteSchemeQuery\x12$.Ydb.Table.ExecuteSchemeQueryRequest\x1a%.Ydb.Table.ExecuteSchemeQueryResponse\x12[\n\x10\x42\x65ginTransaction\x12\".Ydb.Table.BeginTransactionRequest\x1a#.Ydb.Table.BeginTransactionResponse\x12^\n\x11\x43ommitTransaction\x12#.Ydb.Table.CommitTransactionRequest\x1a$.Ydb.Table.CommitTransactionResponse\x12\x64\n\x13RollbackTransaction\x12%.Ydb.Table.RollbackTransactionRequest\x1a&.Ydb.Table.RollbackTransactionResponse\x12g\n\x14\x44\x65scribeTableOptions\x12&.Ydb.Table.DescribeTableOptionsRequest\x1a\'.Ydb.Table.DescribeTableOptionsResponse\x12N\n\x0fStreamReadTable\x12\x1b.Ydb.Table.ReadTableRequest\x1a\x1c.Ydb.Table.ReadTableResponse0\x01\x12\x43\n\x08ReadRows\x12\x1a.Ydb.Table.ReadRowsRequest\x1a\x1b.Ydb.Table.ReadRowsResponse\x12I\n\nBulkUpsert\x12\x1c.Ydb.Table.BulkUpsertRequest\x1a\x1d.Ydb.Table.BulkUpsertResponse\x12j\n\x16StreamExecuteScanQuery\x12\".Ydb.Table.ExecuteScanQueryRequest\x1a*.Ydb.Table.ExecuteScanQueryPartialResponse0\x01\x42O\n\x17tech.ydb.proto.table.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Table_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_table_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027tech.ydb.proto.table.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Table_V1'
  _TABLESERVICE._serialized_start=61
  _TABLESERVICE._serialized_end=1958
# @@protoc_insertion_point(module_scope)



import abc
import json
import logging
import threading
from typing import (
    Iterable,
    Optional,
    Dict,
    Any,
    Union,
)

from . import base
from .base import QueryExplainResultFormat

from .. import _apis, issues, _utilities
from ..settings import BaseRequestSettings
from ..connection import _RpcState as RpcState, EndpointKey
from .._grpc.grpcwrapper import common_utils
from .._grpc.grpcwrapper import ydb_query as _ydb_query
from .._grpc.grpcwrapper import ydb_query_public_types as _ydb_query_public

from .transaction import QueryTxContext

from .._constants import DEFAULT_INITIAL_RESPONSE_TIMEOUT, DEFAULT_LONG_STREAM_TIMEOUT


logger = logging.getLogger(__name__)


def wrapper_create_session(
    rpc_state: RpcState,
    response_pb: _apis.ydb_query.CreateSessionResponse,
    session: "BaseQuerySession",
) -> "BaseQuerySession":
    message = _ydb_query.CreateSessionResponse.from_proto(response_pb)
    issues._process_response(message.status)
    session._session_id = message.session_id
    session._node_id = message.node_id
    return session


def wrapper_delete_session(
    rpc_state: RpcState,
    response_pb: _apis.ydb_query.DeleteSessionResponse,
    session: "BaseQuerySession",
) -> "BaseQuerySession":
    message = _ydb_query.DeleteSessionResponse.from_proto(response_pb)
    issues._process_response(message.status)
    session._closed = True
    return session


class BaseQuerySession(abc.ABC):
    _driver: common_utils.SupportedDriverType
    _settings: base.QueryClientSettings
    _stream = None

    # Session data
    _session_id: Optional[str] = None
    _node_id: Optional[int] = None
    _closed: bool = False

    def __init__(self, driver: common_utils.SupportedDriverType, settings: Optional[base.QueryClientSettings] = None):
        self._driver = driver
        self._settings = self._get_client_settings(driver, settings)
        self._attach_settings: BaseRequestSettings = (
            BaseRequestSettings()
            .with_operation_timeout(DEFAULT_LONG_STREAM_TIMEOUT)
            .with_cancel_after(DEFAULT_LONG_STREAM_TIMEOUT)
            .with_timeout(DEFAULT_LONG_STREAM_TIMEOUT)
        )

        self._last_query_stats = None

    @property
    def session_id(self) -> Optional[str]:
        return self._session_id

    @property
    def node_id(self) -> Optional[int]:
        return self._node_id

    @property
    def is_active(self) -> bool:
        return self._session_id is not None and not self._closed

    @property
    def _endpoint_key(self) -> Optional[EndpointKey]:
        if self._node_id is None:
            return None
        return EndpointKey(endpoint=None, node_id=self._node_id)

    @property
    def is_closed(self) -> bool:
        return self._closed

    @property
    def last_query_stats(self):
        return self._last_query_stats

    def _get_client_settings(
        self,
        driver: common_utils.SupportedDriverType,
        settings: Optional[base.QueryClientSettings] = None,
    ) -> base.QueryClientSettings:
        if settings is not None:
            return settings
        if driver._driver_config.query_client_settings is not None:
            return driver._driver_config.query_client_settings
        return base.QueryClientSettings()

    def _check_session_ready_to_use(self) -> None:
        if not self.is_active:
            raise RuntimeError(f"Session is not active, session_id: {self._session_id}, closed: {self._closed}")

    def _invalidate(self) -> None:
        if self._closed:
            return
        self._closed = True

        if self._stream is not None:
            try:
                self._stream.cancel()
            except Exception:
                pass

    def _on_execute_stream_error(self, e: Exception) -> None:
        if isinstance(e, issues.DeadlineExceed):
            self._invalidate()

    def _create_call(self, settings: Optional[BaseRequestSettings] = None) -> "BaseQuerySession":
        return self._driver(
            _apis.ydb_query.CreateSessionRequest(),
            _apis.QueryService.Stub,
            _apis.QueryService.CreateSession,
            wrap_result=wrapper_create_session,
            wrap_args=(self,),
            settings=settings,
        )

    def _delete_call(self, settings: Optional[BaseRequestSettings] = None) -> "BaseQuerySession":
        return self._driver(
            _apis.ydb_query.DeleteSessionRequest(session_id=self._session_id),
            _apis.QueryService.Stub,
            _apis.QueryService.DeleteSession,
            wrap_result=wrapper_delete_session,
            wrap_args=(self,),
            settings=settings,
            preferred_endpoint=self._endpoint_key,
        )

    def _attach_call(self) -> Iterable[_apis.ydb_query.SessionState]:
        return self._driver(
            _apis.ydb_query.AttachSessionRequest(session_id=self._session_id),
            _apis.QueryService.Stub,
            _apis.QueryService.AttachSession,
            settings=self._attach_settings,
            preferred_endpoint=self._endpoint_key,
        )

    def _execute_call(
        self,
        query: str,
        parameters: dict = None,
        commit_tx: bool = False,
        syntax: base.QuerySyntax = None,
        exec_mode: base.QueryExecMode = None,
        stats_mode: Optional[base.QueryStatsMode] = None,
        schema_inclusion_mode: Optional[base.QuerySchemaInclusionMode] = None,
        result_set_format: Optional[base.QueryResultSetFormat] = None,
        arrow_format_settings: Optional[base.ArrowFormatSettings] = None,
        concurrent_result_sets: bool = False,
        settings: Optional[BaseRequestSettings] = None,
    ) -> Iterable[_apis.ydb_query.ExecuteQueryResponsePart]:
        self._last_query_stats = None

        request = base.create_execute_query_request(
            query=query,
            parameters=parameters,
            commit_tx=commit_tx,
            session_id=self._session_id,
            tx_mode=None,
            tx_id=None,
            syntax=syntax,
            exec_mode=exec_mode,
            stats_mode=stats_mode,
            schema_inclusion_mode=schema_inclusion_mode,
            result_set_format=result_set_format,
            arrow_format_settings=arrow_format_settings,
            concurrent_result_sets=concurrent_result_sets,
        )

        return self._driver(
            request.to_proto(),
            _apis.QueryService.Stub,
            _apis.QueryService.ExecuteQuery,
            settings=settings,
            preferred_endpoint=self._endpoint_key,
        )


class QuerySession(BaseQuerySession):
    """Session object for Query Service. It is not recommended to control
    session's lifecycle manually - use a QuerySessionPool is always a better choise.
    """

    def _attach(self, first_resp_timeout: int = DEFAULT_INITIAL_RESPONSE_TIMEOUT) -> None:
        self._stream = self._attach_call()
        status_stream = _utilities.SyncResponseIterator(
            self._stream,
            lambda response: common_utils.ServerStatus.from_proto(response),
        )

        try:
            first_response = _utilities.get_first_message_with_timeout(
                status_stream,
                first_resp_timeout,
            )
            issues._process_response(first_response)
        except Exception as e:
            self._invalidate()
            raise e

        threading.Thread(
            target=self._check_session_status_loop,
            args=(status_stream,),
            name="attach stream thread",
            daemon=True,
        ).start()

    def _check_session_status_loop(self, status_stream: _utilities.SyncResponseIterator) -> None:
        try:
            for status in status_stream:
                issues._process_response(status)
            logger.debug("Attach stream closed, session_id: %s", self._session_id)
        except Exception as e:
            logger.debug("Attach stream error: %s, session_id: %s", e, self._session_id)
            self._invalidate()

    def delete(self, settings: Optional[BaseRequestSettings] = None) -> None:
        """Deletes a Session of Query Service on server side and releases resources.

        :return: None
        """
        if self._closed:
            return

        if self._session_id:
            try:
                self._delete_call(settings=settings)
            except Exception:
                pass

        self._invalidate()

    def create(self, settings: Optional[BaseRequestSettings] = None) -> "QuerySession":
        """Creates a Session of Query Service on server side and attaches it.

        :return: QuerySession object.
        """
        if self.is_active:
            return self

        if self._closed:
            raise RuntimeError("Session is already closed.")

        self._create_call(settings=settings)
        self._attach()

        return self

    def transaction(self, tx_mode: Optional[base.BaseQueryTxMode] = None) -> QueryTxContext:
        """Creates a transaction context manager with specified transaction mode.

        :param tx_mode: Transaction mode, which is a one from the following choices:
         1) QuerySerializableReadWrite() which is default mode;
         2) QueryOnlineReadOnly(allow_inconsistent_reads=False);
         3) QuerySnapshotReadOnly();
         4) QuerySnapshotReadWrite();
         5) QueryStaleReadOnly().

        :return transaction context manager.

        """
        self._check_session_ready_to_use()

        tx_mode = tx_mode if tx_mode else _ydb_query_public.QuerySerializableReadWrite()

        return QueryTxContext(
            self._driver,
            self,
            tx_mode,
        )

    def execute(
        self,
        query: str,
        parameters: dict = None,
        syntax: base.QuerySyntax = None,
        exec_mode: base.QueryExecMode = None,
        concurrent_result_sets: bool = False,
        settings: Optional[BaseRequestSettings] = None,
        *,
        stats_mode: Optional[base.QueryStatsMode] = None,
        schema_inclusion_mode: Optional[base.QuerySchemaInclusionMode] = None,
        result_set_format: Optional[base.QueryResultSetFormat] = None,
        arrow_format_settings: Optional[base.ArrowFormatSettings] = None,
    ) -> base.SyncResponseContextIterator:
        """Sends a query to Query Service

        :param query: (YQL or SQL text) to be executed.
        :param syntax: Syntax of the query, which is a one from the following choices:
         1) QuerySyntax.YQL_V1, which is default;
         2) QuerySyntax.PG.
        :param parameters: dict with parameters and YDB types;
        :param concurrent_result_sets: A flag to allow YDB mix parts of different result sets. Default is False;
        :param stats_mode: Mode of query statistics to gather, which is a one from the following choices:
         1) QueryStatsMode:NONE, which is default;
         2) QueryStatsMode.BASIC;
         3) QueryStatsMode.FULL;
         4) QueryStatsMode.PROFILE;
        :param schema_inclusion_mode: Schema inclusion mode for result sets:
         1) QuerySchemaInclusionMode.ALWAYS, which is default;
         2) QuerySchemaInclusionMode.FIRST_ONLY.
        :param result_set_format: Format of the result sets:
         1) QueryResultSetFormat.VALUE, which is default;
         2) QueryResultSetFormat.ARROW.
        :param arrow_format_settings: Settings for Arrow format when result_set_format is ARROW.

        :return: Iterator with result sets
        """
        self._check_session_ready_to_use()

        stream_it = self._execute_call(
            query=query,
            parameters=parameters,
            commit_tx=True,
            syntax=syntax,
            exec_mode=exec_mode,
            stats_mode=stats_mode,
            schema_inclusion_mode=schema_inclusion_mode,
            result_set_format=result_set_format,
            arrow_format_settings=arrow_format_settings,
            concurrent_result_sets=concurrent_result_sets,
            settings=settings,
        )

        return base.SyncResponseContextIterator(
            stream_it,
            lambda resp: base.wrap_execute_query_response(
                rpc_state=None,
                response_pb=resp,
                session=self,
                settings=self._settings,
            ),
            on_error=self._on_execute_stream_error,
        )

    def explain(
        self,
        query: str,
        parameters: dict = None,
        *,
        result_format: QueryExplainResultFormat = QueryExplainResultFormat.STR,
    ) -> Union[str, Dict[str, Any]]:
        """Explains query result
        :param query: YQL or SQL query.
        :param parameters: dict with parameters and YDB types;
        :param result_format: Return format: string or dict.
        :return: Parsed query plan.
        """

        res = self.execute(query, parameters, exec_mode=base.QueryExecMode.EXPLAIN)

        # is needs to read result sets for set last_query_stats as sideeffect
        for _ in res:
            pass

        plan = self.last_query_stats.query_plan
        if result_format == QueryExplainResultFormat.DICT:
            plan = json.loads(plan)

        return plan



import pytest
import ydb


@pytest.mark.asyncio
async def test_read_table(driver, database):

    description = (
        ydb.TableDescription()
        .with_primary_keys("key1")
        .with_columns(
            ydb.Column("key1", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("value", ydb.OptionalType(ydb.PrimitiveType.Utf8)),
        )
    )

    session = await driver.table_client.session().create()
    await session.create_table(database + "/test_read_table", description)

    await session.transaction(ydb.SerializableReadWrite()).execute(
        """INSERT INTO `test_read_table` (`key1`, `value`) VALUES (1, "hello_world"), (2, "2")""",
        commit_tx=True,
    )

    expected_res = [[{"key1": 1, "value": "hello_world"}], [{"key1": 2, "value": "2"}]]

    i = 0
    async for resp in await session.read_table(database + "/test_read_table", row_limit=1):
        assert resp.rows == expected_res[i]
        i += 1


@pytest.mark.asyncio
async def test_scan_query_with_query_stats(driver, database):
    description = (
        ydb.TableDescription()
        .with_primary_keys("key1")
        .with_columns(
            ydb.Column("key1", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("value", ydb.OptionalType(ydb.PrimitiveType.Utf8)),
        )
    )

    session = await driver.table_client.session().create()
    await session.create_table(database + "/test_scan_query_with_query_stats", description)
    await session.transaction(ydb.SerializableReadWrite()).execute(
        """INSERT INTO `test_scan_query_with_query_stats` (`key1`, `value`) VALUES (1, "hello_world"), (2, "2")""",
        commit_tx=True,
    )

    expected_res = [{"key1": 1, "value": "hello_world"}, {"key1": 2, "value": "2"}]
    i = 0
    async for resp in await driver.table_client.scan_query(
        "select * from `test_scan_query_with_query_stats` order by key1"
    ):
        for row in resp.result_set.rows:
            assert row == expected_res[i]
            i += 1

    i = 0
    settings = ydb.ScanQuerySettings().with_collect_stats(ydb.QueryStatsCollectionMode.FULL)
    last_response = None
    async for resp in await driver.table_client.scan_query(
        "select * from `test_scan_query_with_query_stats` order by key1",
        settings=settings,
    ):
        last_response = resp
        for row in resp.result_set.rows:
            assert row == expected_res[i]
            i += 1

    assert len(last_response.query_stats.query_phases) > 0


@pytest.mark.asyncio
async def test_read_shard_table(driver, database):
    session: ydb.ISession = await driver.table_client.session().create()
    test_name = "read_shard_table"
    table_name = database + "/" + test_name

    await session.create_table(
        table_name,
        ydb.TableDescription()
        .with_primary_keys("Key1", "Key2")
        .with_columns(
            ydb.Column("Key1", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("Key2", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("Value", ydb.OptionalType(ydb.PrimitiveType.Utf8)),
        )
        .with_profile(ydb.TableProfile().with_partitioning_policy(ydb.PartitioningPolicy().with_uniform_partitions(8))),
    )

    prepared = await session.prepare(
        "DECLARE $data as List<Struct<Key1: Uint64, Key2: Uint64, Value: Utf8>>; "
        "UPSERT INTO {} "
        "SELECT Key1, Key2, Value FROM "
        "AS_TABLE($data);".format(test_name)
    )

    data_by_shard_id = {}
    with session.transaction() as tx:
        max_value = 2**64
        shard_key_bound = max_value >> 3
        data = []

        for shard_id in range(8):
            data_by_shard_id[shard_id] = []

        for idx in range(8 * 8):
            shard_id = idx % 8
            table_row = {
                "Key1": shard_id * shard_key_bound + idx,
                "Key2": idx + 1000,
                "Value": str(idx**4),
            }
            data_by_shard_id[shard_id].append(table_row)
            data.append(table_row)

        await tx.execute(prepared, commit_tx=True, parameters={"$data": data})

    iter = await session.read_table(table_name)
    read_data = [item async for t in iter for item in t.rows]
    assert sorted(read_data, key=lambda x: (x["Key1"], x["Key2"])) == sorted(data, key=lambda x: (x["Key1"], x["Key2"]))



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: protos/ydb_rate_limiter.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'protos/ydb_rate_limiter.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dprotos/ydb_rate_limiter.proto\x12\x0fYdb.RateLimiter\x1a\x1aprotos/ydb_operation.proto\"\x95\x01\n\x17HierarchicalDrrSettings\x12\x1c\n\x14max_units_per_second\x18\x01 \x01(\x01\x12\"\n\x1amax_burst_size_coefficient\x18\x02 \x01(\x01\x12\x1c\n\x14prefetch_coefficient\x18\x03 \x01(\x01\x12\x1a\n\x12prefetch_watermark\x18\x04 \x01(\x01\"o\n\x08Resource\x12\x15\n\rresource_path\x18\x01 \x01(\t\x12\x44\n\x10hierarchical_drr\x18\x02 \x01(\x0b\x32(.Ydb.RateLimiter.HierarchicalDrrSettingsH\x00\x42\x06\n\x04type\"\x9f\x01\n\x15\x43reateResourceRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12+\n\x08resource\x18\x03 \x01(\x0b\x32\x19.Ydb.RateLimiter.Resource\"F\n\x16\x43reateResourceResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x16\n\x14\x43reateResourceResult\"\x9e\x01\n\x14\x41lterResourceRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12+\n\x08resource\x18\x03 \x01(\x0b\x32\x19.Ydb.RateLimiter.Resource\"E\n\x15\x41lterResourceResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x15\n\x13\x41lterResourceResult\"\x87\x01\n\x13\x44ropResourceRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12\x15\n\rresource_path\x18\x03 \x01(\t\"D\n\x14\x44ropResourceResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x14\n\x12\x44ropResourceResult\"\x9b\x01\n\x14ListResourcesRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12\x15\n\rresource_path\x18\x03 \x01(\t\x12\x11\n\trecursive\x18\x04 \x01(\x08\"E\n\x15ListResourcesResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"-\n\x13ListResourcesResult\x12\x16\n\x0eresource_paths\x18\x01 \x03(\t\"\x8b\x01\n\x17\x44\x65scribeResourceRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12\x15\n\rresource_path\x18\x03 \x01(\t\"H\n\x18\x44\x65scribeResourceResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"E\n\x16\x44\x65scribeResourceResult\x12+\n\x08resource\x18\x01 \x01(\x0b\x32\x19.Ydb.RateLimiter.Resource\"\xb7\x01\n\x16\x41\x63quireResourceRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x1e\n\x16\x63oordination_node_path\x18\x02 \x01(\t\x12\x15\n\rresource_path\x18\x03 \x01(\t\x12\x12\n\x08required\x18\x04 \x01(\x04H\x00\x12\x0e\n\x04used\x18\x05 \x01(\x04H\x00\x42\x07\n\x05units\"G\n\x17\x41\x63quireResourceResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x17\n\x15\x41\x63quireResourceResultBu\n\x1btech.ydb.proto.rate_limiterB\x11RateLimiterProtosP\x01Z>github.com/ydb-platform/ydb-go-genproto/protos/Ydb_RateLimiter\xf8\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_rate_limiter_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033tech.ydb.proto.rate_limiterB\021RateLimiterProtosP\001Z>github.com/ydb-platform/ydb-go-genproto/protos/Ydb_RateLimiter\370\001\001'
  _globals['_HIERARCHICALDRRSETTINGS']._serialized_start=79
  _globals['_HIERARCHICALDRRSETTINGS']._serialized_end=228
  _globals['_RESOURCE']._serialized_start=230
  _globals['_RESOURCE']._serialized_end=341
  _globals['_CREATERESOURCEREQUEST']._serialized_start=344
  _globals['_CREATERESOURCEREQUEST']._serialized_end=503
  _globals['_CREATERESOURCERESPONSE']._serialized_start=505
  _globals['_CREATERESOURCERESPONSE']._serialized_end=575
  _globals['_CREATERESOURCERESULT']._serialized_start=577
  _globals['_CREATERESOURCERESULT']._serialized_end=599
  _globals['_ALTERRESOURCEREQUEST']._serialized_start=602
  _globals['_ALTERRESOURCEREQUEST']._serialized_end=760
  _globals['_ALTERRESOURCERESPONSE']._serialized_start=762
  _globals['_ALTERRESOURCERESPONSE']._serialized_end=831
  _globals['_ALTERRESOURCERESULT']._serialized_start=833
  _globals['_ALTERRESOURCERESULT']._serialized_end=854
  _globals['_DROPRESOURCEREQUEST']._serialized_start=857
  _globals['_DROPRESOURCEREQUEST']._serialized_end=992
  _globals['_DROPRESOURCERESPONSE']._serialized_start=994
  _globals['_DROPRESOURCERESPONSE']._serialized_end=1062
  _globals['_DROPRESOURCERESULT']._serialized_start=1064
  _globals['_DROPRESOURCERESULT']._serialized_end=1084
  _globals['_LISTRESOURCESREQUEST']._serialized_start=1087
  _globals['_LISTRESOURCESREQUEST']._serialized_end=1242
  _globals['_LISTRESOURCESRESPONSE']._serialized_start=1244
  _globals['_LISTRESOURCESRESPONSE']._serialized_end=1313
  _globals['_LISTRESOURCESRESULT']._serialized_start=1315
  _globals['_LISTRESOURCESRESULT']._serialized_end=1360
  _globals['_DESCRIBERESOURCEREQUEST']._serialized_start=1363
  _globals['_DESCRIBERESOURCEREQUEST']._serialized_end=1502
  _globals['_DESCRIBERESOURCERESPONSE']._serialized_start=1504
  _globals['_DESCRIBERESOURCERESPONSE']._serialized_end=1576
  _globals['_DESCRIBERESOURCERESULT']._serialized_start=1578
  _globals['_DESCRIBERESOURCERESULT']._serialized_end=1647
  _globals['_ACQUIRERESOURCEREQUEST']._serialized_start=1650
  _globals['_ACQUIRERESOURCEREQUEST']._serialized_end=1833
  _globals['_ACQUIRERESOURCERESPONSE']._serialized_start=1835
  _globals['_ACQUIRERESOURCERESPONSE']._serialized_end=1906
  _globals['_ACQUIRERESOURCERESULT']._serialized_start=1908
  _globals['_ACQUIRERESOURCERESULT']._serialized_end=1931
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_query_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2
from ydb._grpc.v3.protos import ydb_query_pb2 as protos_dot_ydb__query__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_query_v1.proto',
  package='Ydb.Query.V1',
  syntax='proto3',
  serialized_options=b'\n\027tech.ydb.proto.query.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Query_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12ydb_query_v1.proto\x12\x0cYdb.Query.V1\x1a\x1aprotos/ydb_operation.proto\x1a\x16protos/ydb_query.proto2\xad\x06\n\x0cQueryService\x12R\n\rCreateSession\x12\x1f.Ydb.Query.CreateSessionRequest\x1a .Ydb.Query.CreateSessionResponse\x12R\n\rDeleteSession\x12\x1f.Ydb.Query.DeleteSessionRequest\x1a .Ydb.Query.DeleteSessionResponse\x12K\n\rAttachSession\x12\x1f.Ydb.Query.AttachSessionRequest\x1a\x17.Ydb.Query.SessionState0\x01\x12[\n\x10\x42\x65ginTransaction\x12\".Ydb.Query.BeginTransactionRequest\x1a#.Ydb.Query.BeginTransactionResponse\x12^\n\x11\x43ommitTransaction\x12#.Ydb.Query.CommitTransactionRequest\x1a$.Ydb.Query.CommitTransactionResponse\x12\x64\n\x13RollbackTransaction\x12%.Ydb.Query.RollbackTransactionRequest\x1a&.Ydb.Query.RollbackTransactionResponse\x12U\n\x0c\x45xecuteQuery\x12\x1e.Ydb.Query.ExecuteQueryRequest\x1a#.Ydb.Query.ExecuteQueryResponsePart0\x01\x12K\n\rExecuteScript\x12\x1f.Ydb.Query.ExecuteScriptRequest\x1a\x19.Ydb.Operations.Operation\x12\x61\n\x12\x46\x65tchScriptResults\x12$.Ydb.Query.FetchScriptResultsRequest\x1a%.Ydb.Query.FetchScriptResultsResponseBO\n\x17tech.ydb.proto.query.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Query_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__operation__pb2.DESCRIPTOR,protos_dot_ydb__query__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_QUERYSERVICE = _descriptor.ServiceDescriptor(
  name='QueryService',
  full_name='Ydb.Query.V1.QueryService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=89,
  serialized_end=902,
  methods=[
  _descriptor.MethodDescriptor(
    name='CreateSession',
    full_name='Ydb.Query.V1.QueryService.CreateSession',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._CREATESESSIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._CREATESESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteSession',
    full_name='Ydb.Query.V1.QueryService.DeleteSession',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._DELETESESSIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._DELETESESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AttachSession',
    full_name='Ydb.Query.V1.QueryService.AttachSession',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._ATTACHSESSIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._SESSIONSTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BeginTransaction',
    full_name='Ydb.Query.V1.QueryService.BeginTransaction',
    index=3,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._BEGINTRANSACTIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._BEGINTRANSACTIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CommitTransaction',
    full_name='Ydb.Query.V1.QueryService.CommitTransaction',
    index=4,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._COMMITTRANSACTIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._COMMITTRANSACTIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RollbackTransaction',
    full_name='Ydb.Query.V1.QueryService.RollbackTransaction',
    index=5,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._ROLLBACKTRANSACTIONREQUEST,
    output_type=protos_dot_ydb__query__pb2._ROLLBACKTRANSACTIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExecuteQuery',
    full_name='Ydb.Query.V1.QueryService.ExecuteQuery',
    index=6,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._EXECUTEQUERYREQUEST,
    output_type=protos_dot_ydb__query__pb2._EXECUTEQUERYRESPONSEPART,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExecuteScript',
    full_name='Ydb.Query.V1.QueryService.ExecuteScript',
    index=7,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._EXECUTESCRIPTREQUEST,
    output_type=protos_dot_ydb__operation__pb2._OPERATION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='FetchScriptResults',
    full_name='Ydb.Query.V1.QueryService.FetchScriptResults',
    index=8,
    containing_service=None,
    input_type=protos_dot_ydb__query__pb2._FETCHSCRIPTRESULTSREQUEST,
    output_type=protos_dot_ydb__query__pb2._FETCHSCRIPTRESULTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_QUERYSERVICE)

DESCRIPTOR.services_by_name['QueryService'] = _QUERYSERVICE

# @@protoc_insertion_point(module_scope)



import asyncio

from .reconnector import CoordinationReconnector
from .semaphore import CoordinationSemaphore


class CoordinationSession:
    def __init__(self, driver, path: str, timeout_millis: int = 30000):
        self._driver = driver
        self._path = path
        self._timeout_millis = timeout_millis

        self._reconnector = CoordinationReconnector(
            driver=driver,
            node_path=path,
            timeout_millis=timeout_millis,
        )

        self._req_id = 0
        self._req_id_lock = asyncio.Lock()
        self._closed = False

    async def next_req_id(self) -> int:
        async with self._req_id_lock:
            self._req_id += 1
            return self._req_id

    def semaphore(self, name: str, limit: int = 1) -> CoordinationSemaphore:
        if self._closed:
            raise RuntimeError("CoordinationSession is closed")
        return CoordinationSemaphore(self, name, limit)

    async def close(self):
        if self._closed:
            return
        self._closed = True
        await self._reconnector.stop()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_import_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_import_pb2 as protos_dot_ydb__import__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_import_v1.proto',
  package='Ydb.Import.V1',
  syntax='proto3',
  serialized_options=b'\n\031tech.ydb.proto.import_.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Import_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13ydb_import_v1.proto\x12\rYdb.Import.V1\x1a\x17protos/ydb_import.proto2\xde\x03\n\rImportService\x12Q\n\x0cImportFromS3\x12\x1f.Ydb.Import.ImportFromS3Request\x1a .Ydb.Import.ImportFromS3Response\x12Q\n\x0cImportFromFs\x12\x1f.Ydb.Import.ImportFromFsRequest\x1a .Ydb.Import.ImportFromFsResponse\x12l\n\x15ListObjectsInS3Export\x12(.Ydb.Import.ListObjectsInS3ExportRequest\x1a).Ydb.Import.ListObjectsInS3ExportResponse\x12l\n\x15ListObjectsInFsExport\x12(.Ydb.Import.ListObjectsInFsExportRequest\x1a).Ydb.Import.ListObjectsInFsExportResponse\x12K\n\nImportData\x12\x1d.Ydb.Import.ImportDataRequest\x1a\x1e.Ydb.Import.ImportDataResponseBR\n\x19tech.ydb.proto.import_.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Import_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__import__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_IMPORTSERVICE = _descriptor.ServiceDescriptor(
  name='ImportService',
  full_name='Ydb.Import.V1.ImportService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=64,
  serialized_end=542,
  methods=[
  _descriptor.MethodDescriptor(
    name='ImportFromS3',
    full_name='Ydb.Import.V1.ImportService.ImportFromS3',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__import__pb2._IMPORTFROMS3REQUEST,
    output_type=protos_dot_ydb__import__pb2._IMPORTFROMS3RESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ImportFromFs',
    full_name='Ydb.Import.V1.ImportService.ImportFromFs',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__import__pb2._IMPORTFROMFSREQUEST,
    output_type=protos_dot_ydb__import__pb2._IMPORTFROMFSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListObjectsInS3Export',
    full_name='Ydb.Import.V1.ImportService.ListObjectsInS3Export',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__import__pb2._LISTOBJECTSINS3EXPORTREQUEST,
    output_type=protos_dot_ydb__import__pb2._LISTOBJECTSINS3EXPORTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListObjectsInFsExport',
    full_name='Ydb.Import.V1.ImportService.ListObjectsInFsExport',
    index=3,
    containing_service=None,
    input_type=protos_dot_ydb__import__pb2._LISTOBJECTSINFSEXPORTREQUEST,
    output_type=protos_dot_ydb__import__pb2._LISTOBJECTSINFSEXPORTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ImportData',
    full_name='Ydb.Import.V1.ImportService.ImportData',
    index=4,
    containing_service=None,
    input_type=protos_dot_ydb__import__pb2._IMPORTDATAREQUEST,
    output_type=protos_dot_ydb__import__pb2._IMPORTDATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_IMPORTSERVICE)

DESCRIPTOR.services_by_name['ImportService'] = _IMPORTSERVICE

# @@protoc_insertion_point(module_scope)
